from airtouch5py.packets.datapacket import Data


class AcErrorInformationRequestData(Data):
    ac_number: int

    def __init__(self, ac_number: int):
        self.ac_number = ac_number


class AcErrorInformationData(Data):
    ac_number: int
    error_info: str

    def __init__(self, ac_number: int, error_info: str):
        self.ac_number = ac_number
        self.error_info = error_info



import asyncio
import binascii
import logging
from enum import Enum

from airtouch5py.data_packet_factory import DataPacketFactory

from airtouch5py.packet_encoder import PacketEncoder
from airtouch5py.packet_reader import PacketReader

from airtouch5py.packets.datapacket import DataPacket

_LOGGER = logging.getLogger(__name__)


class Airtouch5ConnectionStateChange(Enum):
    CONNECTED = 1
    DISCONNECTED = 2


class Airtouch5Client:
    """
    The "Raw" (Hard mode) Airtouch 5 Client.
    You should probably use Airtouch5SimpleClient instead.

    Usage:
    Construct, call connect.
    Wait on packets_received to receive a CONNECTED message.
    Wait on updates from packets_received, use send_packet to send packets.

    Call disconnect to disconnect.

    When a DISCONNECTED message is received, call connect to reconnect.
    """

    _encoder = PacketEncoder()
    _packet_reader = PacketReader()

    ip: str
    packets_received: asyncio.Queue[Airtouch5ConnectionStateChange | DataPacket]

    _reader: asyncio.StreamReader | None
    _writer: asyncio.StreamWriter | None

    _disconnect_lock: asyncio.Lock

    def __init__(self, ip: str):
        self.ip = ip
        self.packets_received = asyncio.Queue()
        self.data_packet_factory = DataPacketFactory()

        self._connected = False
        self._should_be_connected = False
        self._writer, self._reader = None, None
        self._disconnect_lock = asyncio.Lock()

    async def connect(self):
        """
        Connect to the airtouch 5.
        Times out after 5 seconds.
        Throws if we fail to connect.
        Otherwise puts a connected message in the queue and starts up the reader task.
        """

        # Clear the queue before we connect (Might be dangling stuff in it from a previous connection)
        while not self.packets_received.empty():
            self.packets_received.get_nowait()

        _LOGGER.info(f"Connecting to {self.ip}:9005")
        self._reader, self._writer = await asyncio.wait_for(
            asyncio.open_connection(self.ip, 9005), 5
        )
        _LOGGER.info(f"Connected to {self.ip}:9005")

        self.packets_received.put_nowait(Airtouch5ConnectionStateChange.CONNECTED)
        self._reader_task = asyncio.create_task(self._read_packets())

    async def disconnect(self):
        """
        Disconnect the socket if it is connected.
        If stop_reader_task is true, also stop the reader task (Normally you should do this, unless called from the reader task)
        Pushes a DISCONNECTED message in to the queue if we actually disconnected (were connected before)
        """
        async with self._disconnect_lock:
            did_disconnect = False
            if self._writer is not None:
                did_disconnect = True
                self._writer.close()
                try:
                    await self._writer.wait_closed()
                except Exception:
                    # Ignore exceptions when closing
                    _LOGGER.debug("Exception when closing writer", exc_info=True)

            if self._reader_task is not None:
                self._reader_task.cancel()
                self._reader_task = None

            self._writer, self._reader = None, None
            if did_disconnect:
                self.packets_received.put_nowait(
                    Airtouch5ConnectionStateChange.DISCONNECTED
                )

    async def _read_packets(self):
        """
        Continuously read packets from the socket and put them in the queue.
        If we detect the socket has died, disconnect it
        """
        reader = self._reader
        if reader is None:
            raise Exception("Reader is None")

        try:
            while reader.at_eof() == False:
                read = await reader.read(1024)
                _LOGGER.debug(f"Received data: {binascii.hexlify(read)}")
                packets = self._packet_reader.read(read)
                for packet in packets:
                    self.packets_received.put_nowait(packet)
        except Exception as e:
            _LOGGER.error(f"Exception in reader task: {e}")

        # If we've finished reading for some reason, we should disconnect
        # Clear our task first so we don't get cancelled
        self._reader_task = None
        await self.disconnect()

    async def send_packet(self, packet: DataPacket):
        """
        Send the given packet to the airtouch 5.
        Throws if we aren't connected or if there is a connection issue
        """
        writer = self._writer
        if writer is None:
            raise Exception("Writer is None")

        try:
            data = self._encoder.encode(packet)
            _LOGGER.debug(f"Sending data: {binascii.hexlify(data)}")
            writer.write(data)
            await writer.drain()
        except Exception as e:
            _LOGGER.error(f"Exception when sending packet: {e}")
            await self.disconnect()
            raise e



import asyncio
import logging
from typing import Callable, TypeVar

from airtouch5py.airtouch5_client import Airtouch5Client, Airtouch5ConnectionStateChange
from airtouch5py.data_packet_factory import DataPacketFactory
from airtouch5py.packets.ac_ability import AcAbility, AcAbilityData
from airtouch5py.packets.ac_status import AcStatus, AcStatusData
from airtouch5py.packets.console_version import ConsoleVersionData
from airtouch5py.packets.datapacket import Data, DataPacket
from airtouch5py.packets.zone_name import ZoneName, ZoneNameData
from airtouch5py.packets.zone_status import ZoneStatusData, ZoneStatusZone

_LOGGER = logging.getLogger(__name__)
T = TypeVar("T")


class Airtouch5SimpleClient:
    """
    A simple Airtouch5 client.

    Usage:
    Call connect_and_stay_connected().
    Add listeners to *_callbacks
    The Airtouch5 will automatically send out updates to zone status and ac status as they happen.
    """

    data_packet_factory: DataPacketFactory

    # Populated after connect_and_stay_connected
    ac: list[AcAbility]
    # Populated after connect_and_stay_connected
    zones: list[ZoneName]
    # Populated after connect_and_stay_connected
    console_version: str
    # Populated after connect_and_stay_connected
    latest_ac_status: dict[int, AcStatus]
    # Populated after connect_and_stay_connected
    latest_zone_status: dict[int, ZoneStatusZone]

    connection_state_callbacks: list[Callable[[Airtouch5ConnectionStateChange], None]]
    data_packet_callbacks: list[Callable[[DataPacket], None]]
    ac_status_callbacks: list[Callable[[dict[int, AcStatus]], None]]
    zone_status_callbacks: list[Callable[[dict[int, ZoneStatusZone]], None]]

    _client: Airtouch5Client
    _connection_task: asyncio.Task[None] | None

    def __init__(self, ip: str):
        self.ip = ip
        self._client = Airtouch5Client(ip)
        self.data_packet_factory = DataPacketFactory()

        self.ac = []
        self.zones = []
        self.console_version = ""

        self.connection_state_callbacks = []
        self.data_packet_callbacks = []
        self.ac_status_callbacks = []
        self.zone_status_callbacks = []

    async def test_connection(self) -> None:
        """
        Connect, verify the connection, disconnect.
        Throws if something goes wrong.
        """
        await self._client.connect()

        try:
            # Send a console version request to verify this is an Airtouch 5 console
            await self._client.send_packet(
                self.data_packet_factory.console_version_request()
            )

            # Wait for the response
            start_wait = asyncio.get_running_loop().time()
            got_response = False
            while (
                asyncio.get_running_loop().time() - start_wait < 5 and not got_response
            ):
                packet = await asyncio.wait_for(self._client.packets_received.get(), 5)
                if isinstance(packet, DataPacket) and isinstance(
                    packet.data, ConsoleVersionData
                ):
                    got_response = True
                    break

            if not got_response:
                raise Exception("Didn't receive a console version response")
        finally:
            await self._client.disconnect()

    async def connect_and_stay_connected(self) -> None:
        """
        Connect, and reconnect if we disconnect.
        Gets the AC ability and zone names, initial zone status and ac status, and then waits for updates.
        Throws if we fail to make the initial connection.
        """
        await self._client.connect()

        # Get the ac abilities
        await self._client.send_packet(self.data_packet_factory.ac_ability_request())
        self.ac = (await self._wait_for_packet_or_throw(AcAbilityData)).ac_ability

        # Get the zone names
        await self._client.send_packet(self.data_packet_factory.zone_name_request())
        self.zones = (await self._wait_for_packet_or_throw(ZoneNameData)).zone_names

        # Get the version
        await self._client.send_packet(
            self.data_packet_factory.console_version_request()
        )
        self.console_version = (
            await self._wait_for_packet_or_throw(ConsoleVersionData)
        ).version

        # Get the initial zone status
        await self._client.send_packet(self.data_packet_factory.zone_status_request())
        self.latest_zone_status = {
            zone.zone_number: zone
            for zone in (await self._wait_for_packet_or_throw(ZoneStatusData)).zones
        }

        # Get the initial ac status
        await self._client.send_packet(self.data_packet_factory.ac_status_request())
        self.latest_ac_status = {
            ac.ac_number: ac
            for ac in (await self._wait_for_packet_or_throw(AcStatusData)).ac_status
        }

        # Start up the connection/reader task
        self._connection_task = asyncio.create_task(self._maintain_connection())

    async def _wait_for_packet_or_throw(self, packet_type: type[T]) -> T:
        """
        Wait 5 seconds for a packet of the given type, or throw if we disconnect or timeout.
        """

        async def _read_packets_until_match() -> T:
            while True:
                packet = await self._client.packets_received.get()
                if packet is Airtouch5ConnectionStateChange.DISCONNECTED:
                    raise Exception(f"Disconnected while waiting for a {packet_type}")
                if isinstance(packet, DataPacket) and isinstance(
                    packet.data, packet_type
                ):
                    return packet.data

        return await asyncio.wait_for(_read_packets_until_match(), 5)

    async def _maintain_connection(self) -> None:
        """
        Read messages off the queue, reconnecting if we disconnect.
        Calls the matching callbacks.
        """

        # AirTouch5 doesn't send any packets if nothing is changing.
        # So we periodically send a packet to test if the connection is alive.
        have_sent_keep_alive_packet = False

        while True:
            # Wait for a packet, send one after 4 minutes, or timeout after 4+1 minutes
            packet: DataPacket | Airtouch5ConnectionStateChange
            try:
                packet = await asyncio.wait_for(
                    self._client.packets_received.get(),
                    1 * 60 if have_sent_keep_alive_packet else 4 * 60,
                )
            except asyncio.TimeoutError:
                if not have_sent_keep_alive_packet:
                    # send something to test the connection
                    try:
                        await self._client.send_packet(
                            self.data_packet_factory.console_version_request()
                        )
                        have_sent_keep_alive_packet = True
                    except:
                        # Ignore, send_packet will disconnect if it fails
                        _LOGGER.info(
                            "Failed to send keep alive packet, connection must be dead"
                        )
                    continue
                else:
                    _LOGGER.error("Timeout waiting for packet, reconnecting")
                    await self._client.disconnect()
                    have_sent_keep_alive_packet = False
                    # disconnect pushes a DISCONNECTED message in to the queue, so we'll reconnect
                    continue

            _LOGGER.debug(f"maintain Received packet {packet}")
            have_sent_keep_alive_packet = False

            if packet is Airtouch5ConnectionStateChange.DISCONNECTED:
                [cb(packet) for cb in self.connection_state_callbacks]
                _LOGGER.warning("Disconnected from Airtouch 5, reconnecting")
                while True:
                    try:
                        await self._client.connect()
                        break
                    except Exception as e:
                        _LOGGER.error(
                            f"Failed to reconnect: {e}, will reconnect in 5 seconds"
                        )
                        await asyncio.sleep(5)
            elif packet is Airtouch5ConnectionStateChange.CONNECTED:
                [cb(packet) for cb in self.connection_state_callbacks]
            elif isinstance(packet, DataPacket):
                [cb(packet) for cb in self.data_packet_callbacks]
                if isinstance(packet.data, ZoneStatusData):
                    # convert the list to a dict, store it and broadcast it
                    self.latest_zone_status = {
                        zone.zone_number: zone for zone in packet.data.zones
                    }
                    [cb(self.latest_zone_status) for cb in self.zone_status_callbacks]
                if isinstance(packet.data, AcStatusData):
                    # convert the list to a dict, store it and broadcast it
                    self.latest_ac_status = {
                        ac.ac_number: ac for ac in packet.data.ac_status
                    }
                    [cb(self.latest_ac_status) for cb in self.ac_status_callbacks]
            else:
                _LOGGER.error(f"Received unknown packet type {packet}")

    async def send_packet(self, packet: DataPacket) -> None:
        """
        Send a packet.
        """
        await self._client.send_packet(packet)

    async def disconnect(self) -> None:
        """
        Disconnect, and stop reconnecting.
        """
        if self._connection_task is not None:
            self._connection_task.cancel()
        await self._client.disconnect()



from airtouch5py.packets.datapacket import Data


class AcAbility:
    ac_number: int
    ac_name: str
    start_zone_number: int
    zone_count: int

    supports_mode_cool: bool
    supports_mode_fan: bool
    supports_mode_dry: bool
    supports_mode_heat: bool
    supports_mode_auto: bool

    supports_fan_speed_intelligent_auto: bool
    supports_fan_speed_turbo: bool
    supports_fan_speed_powerful: bool
    supports_fan_speed_high: bool
    supports_fan_speed_medium: bool
    supports_fan_speed_low: bool
    supports_fan_speed_quiet: bool  # quite in docs
    supports_fan_speed_auto: bool

    # The docs are wrong for these (they are all cool)
    min_cool_set_point: int
    max_cool_set_point: int
    min_heat_set_point: int
    max_heat_set_point: int

    def __init__(
        self,
        ac_number: int,
        ac_name: str,
        start_zone_number: int,
        zone_count: int,
        supports_mode_cool: bool,
        supports_mode_fan: bool,
        supports_mode_dry: bool,
        supports_mode_heat: bool,
        supports_mode_auto: bool,
        supports_fan_speed_intelligent_auto: bool,
        supports_fan_speed_turbo: bool,
        supports_fan_speed_powerful: bool,
        supports_fan_speed_high: bool,
        supports_fan_speed_medium: bool,
        supports_fan_speed_low: bool,
        supports_fan_speed_quiet: bool,
        supports_fan_speed_auto: bool,
        min_cool_set_point: int,
        max_cool_set_point: int,
        min_heat_set_point: int,
        max_heat_set_point: int,
    ):
        self.ac_number = ac_number
        self.ac_name = ac_name
        self.start_zone_number = start_zone_number
        self.zone_count = zone_count
        self.supports_mode_cool = supports_mode_cool
        self.supports_mode_fan = supports_mode_fan
        self.supports_mode_dry = supports_mode_dry
        self.supports_mode_heat = supports_mode_heat
        self.supports_mode_auto = supports_mode_auto
        self.supports_fan_speed_intelligent_auto = supports_fan_speed_intelligent_auto
        self.supports_fan_speed_turbo = supports_fan_speed_turbo
        self.supports_fan_speed_powerful = supports_fan_speed_powerful
        self.supports_fan_speed_high = supports_fan_speed_high
        self.supports_fan_speed_medium = supports_fan_speed_medium
        self.supports_fan_speed_low = supports_fan_speed_low
        self.supports_fan_speed_quiet = supports_fan_speed_quiet
        self.supports_fan_speed_auto = supports_fan_speed_auto
        self.min_cool_set_point = min_cool_set_point
        self.max_cool_set_point = max_cool_set_point
        self.min_heat_set_point = min_heat_set_point
        self.max_heat_set_point = max_heat_set_point


class AcAbilityData(Data):
    ac_ability: list[AcAbility]

    def __init__(self, ac_ability: list[AcAbility]):
        self.ac_ability = ac_ability


class AcAbilityRequestData(Data):
    ac_number: int | None  # None to request the ability of all ACs

    def __init__(self, ac_number: int | None):
        self.ac_number = ac_number



import struct

from airtouch5py.packet_fields import (
    ControlStatusSubType,
    ExtendedMessageSubType,
    MessageType,
)

from airtouch5py.packets.ac_ability import (
    AcAbility,
    AcAbilityData,
    AcAbilityRequestData,
)

from airtouch5py.packets.ac_control import (
    AcControl,
    AcControlData,
    SetAcFanSpeed,
    SetAcMode,
    SetpointControl,
    SetPowerSetting,
)
from airtouch5py.packets.ac_error_information import (
    AcErrorInformationData,
    AcErrorInformationRequestData,
)
from airtouch5py.packets.ac_status import (
    AcFanSpeed,
    AcMode,
    AcPowerState,
    AcStatus,
    AcStatusData,
)
from airtouch5py.packets.console_version import (
    ConsoleVersionData,
    ConsoleVersionRequestData,
)

from airtouch5py.packets.datapacket import Data, DataPacket
from airtouch5py.packets.zone_control import (
    ZoneControlData,
    ZoneControlZone,
    ZoneSettingPower,
    ZoneSettingValue,
)
from airtouch5py.packets.zone_name import ZoneName, ZoneNameData, ZoneNameRequestData

from airtouch5py.packets.zone_status import (
    ControlMethod,
    ZonePowerState,
    ZoneStatusData,
    ZoneStatusZone,
)
from bitarray import bitarray
from bitarray.util import ba2int


class PacketDecoder:
    """
    Decode packets from the AirTouch 5 protocol.
    Assumes that they have already been validated (CRC, data length)
    """

    # https://stackoverflow.com/questions/43634618/how-do-i-test-if-int-value-exists-in-python-enum-without-using-try-catch
    _set_ZoneSettingValue = set(item.value for item in ZoneSettingValue)
    _set_ZoneSettingPower = set(item.value for item in ZoneSettingPower)
    _set_SetPowerSetting = set(item.value for item in SetPowerSetting)
    _set_SetAcMode = set(item.value for item in SetAcMode)
    _set_SetAcFanSpeed = set(item.value for item in SetAcFanSpeed)
    _set_SetpointControl = set(item.value for item in SetpointControl)
    _set_AcPowerState = set(item.value for item in AcPowerState)
    _set_AcMode = set(item.value for item in AcMode)
    _set_AcFanSpeed = set(item.value for item in AcFanSpeed)

    def decode(self, buffer: bytes) -> DataPacket:
        # Header (4 bytes)
        # Address (2 bytes)
        # Message ID (1 byte)
        # Message type (1 byte)
        # Data length (2 bytes)
        # Data (variable)
        # CRC (2 bytes)

        # if len(bytes) < 4 + 2 + 1 + 1 + 2 + 2:
        #    raise ValueError("Packet is too short")

        # TODO: Could do these all in one unpack call?
        address: int = struct.unpack(">H", buffer[4:6])[0]
        message_id: int = struct.unpack(">B", buffer[6:7])[0]
        message_type: int = struct.unpack(">B", buffer[7:8])[0]
        data_length: int = struct.unpack(">H", buffer[8:10])[0]
        data_bytes: bytes = buffer[10 : data_length + 10]

        data: Data
        match message_type:
            case MessageType.CONTROL_STATUS.value:
                data = self.decode_control_status(data_bytes)
            case MessageType.EXTENDED.value:
                data = self.decode_extended(data_bytes)
            case _:
                raise ValueError(f"Unknown message type: {hex(message_type)}")

        return DataPacket(address, message_id, data)

    def decode_control_status(self, bytes: bytes) -> Data:
        # Sub message type (1 byte)
        # Keep 0 (1 byte)
        # Normal data length (2 bytes)
        # Each repeat data length (2 bytes)
        # Repeat data count (2 bytes)

        sub_message_type = struct.unpack(">B", bytes[0:1])[0]
        normal_data_length = struct.unpack(">H", bytes[2:4])[0]
        repeat_data_length = struct.unpack(">H", bytes[4:6])[0]
        repeat_data_count = struct.unpack(">H", bytes[6:8])[0]

        # Should now have normal_data_length of data, then repeat length * data_count
        if (
            len(bytes) - 8
            != normal_data_length + repeat_data_length * repeat_data_count
        ):
            raise ValueError(
                f"Data length does not match expected {normal_data_length + repeat_data_length * repeat_data_count} found {len(bytes) - 8}"
            )

        match sub_message_type:
            case ControlStatusSubType.ZONE_CONTROL.value:
                if normal_data_length != 0:
                    raise ValueError("Zone control message should not have normal data")
                if repeat_data_length != 4:
                    raise ValueError(
                        "Zone control message should have 4 byte repeat data"
                    )
                return self.decode_zone_control(bytes[8 + normal_data_length:], repeat_data_count)
            case ControlStatusSubType.ZONE_STATUS.value:
                #This packet doesn't currently have normal data (1.2.0), but the docs say it might in the future
                if repeat_data_count != 0 and repeat_data_length < 8:
                    raise ValueError(
                        "Zone status message should have at least 8 byte repeat data"
                    )
                return self.decode_zone_status(bytes[8 + normal_data_length:], repeat_data_count, repeat_data_length)
            case ControlStatusSubType.AC_CONTROL.value:
                if normal_data_length != 0:
                    raise ValueError("AC control message should not have normal data")
                if repeat_data_length != 4:
                    raise ValueError(
                        "AC control message should have 4 byte repeat data"
                    )
                return self.decode_ac_control(bytes[8 + normal_data_length:], repeat_data_count)
            case ControlStatusSubType.AC_STATUS.value:
                #This packet doesn't currently have normal data (1.2.0), but the docs say it might in the future
                if repeat_data_count != 0 and repeat_data_length < 10:
                    raise ValueError(
                        f"AC status message should have at least 10 or 14 byte repeat data, but it was {repeat_data_length}"
                    )
                return self.decode_ac_status(
                    bytes[8 + normal_data_length:], repeat_data_count, repeat_data_length
                )
            case _:
                raise ValueError(f"Unknown sub message type: {hex(sub_message_type)}")

    def decode_zone_control(
        self, bytes: bytes, repeat_data_count: int
    ) -> ZoneControlData:
        zones: list[ZoneControlZone] = []
        bits = bitarray(endian="big")

        for i in range(0, repeat_data_count):
            bits.clear()
            bits.frombytes(bytes[i * 4 : i * 4 + 4])
            # Byte 1 Bit 6-1 Zone number
            zone_number = ba2int(bits[2:8])
            # Byte 2 Bit 8-6 Zone setting value (Can be invalid -> KEEP_SETTING_VALUE)
            zone_setting_value = ba2int(bits[8 + 0 : 8 + 3])
            if zone_setting_value in self._set_ZoneSettingValue:
                zone_setting_value = ZoneSettingValue(zone_setting_value)
            else:
                zone_setting_value = ZoneSettingValue.KEEP_SETTING_VALUE
            # Byte 2 bit 3-1 Power
            power = ba2int(bits[8 + 5 : 8 + 8])
            if power in self._set_ZoneSettingPower:
                power = ZoneSettingPower(power)
            else:
                power = ZoneSettingPower.KEEP_POWER_STATE
            # Byte 3 Value to set
            value_to_set = ba2int(bits[16 : 16 + 8])

            if zone_setting_value == ZoneSettingValue.SET_OPEN_PERCENTAGE:
                value_to_set = value_to_set / 100
            elif zone_setting_value == ZoneSettingValue.SET_TARGET_SETPOINT:
                value_to_set = (value_to_set + 100) / 10

            zones.append(
                ZoneControlZone(zone_number, zone_setting_value, power, value_to_set)
            )
        return ZoneControlData(zones)

    def decode_zone_status(
        self, bytes: bytes, repeat_data_count: int, repeat_data_length: int
    ) -> ZoneStatusData:
        if repeat_data_count == 0:
            return ZoneStatusData([])

        zones: list[ZoneStatusZone] = []
        bits = bitarray(endian="big")

        for i in range(0, repeat_data_count):
            bits.clear()
            bits.frombytes(bytes[i * repeat_data_length : i * repeat_data_length + repeat_data_length])
            # Byte 1 Bit 8-7 Zone power state
            zone_power_state = ZonePowerState(ba2int(bits[0:2]))
            # Byte 1 Bit 6-1 Zone number
            zone_number = ba2int(bits[2:8])
            # Byte 2 Bit 8 Control method
            control_method = ControlMethod(ba2int(bits[8 + 0 : 8 + 1]))
            # Byte 2 Bit 7-1 Open percentage
            open_percentage = ba2int(bits[8 + 1 : 8 + 8]) / 100
            # Byte 3 Set point
            set_point = ba2int(bits[16 : 16 + 8])
            if set_point == 0xFF:  # 0xFF invalid
                set_point = None
            else:
                set_point = (set_point + 100) / 10
            # Byte 4 Bit 8 Has sensor
            has_sensor = bool(bits[24 + 0])
            # Byte 5 Bit 3-1, Byte 6 Temperature
            temperature = ba2int(bits[32 + 5 : 32 + 5 + 3 + 8])
            if temperature <= 2000:
                temperature = (temperature - 500) / 10
            else:  # Other: Not available
                temperature = None
            # Byte 7 Bit 2 Spill active
            spill_active = bool(bits[48 + 6])
            # Byte 7 Bit 1 Is low battery
            is_low_battery = bool(bits[48 + 7])

            zones.append(
                ZoneStatusZone(
                    zone_power_state,
                    zone_number,
                    control_method,
                    open_percentage,
                    set_point,
                    has_sensor,
                    temperature,
                    spill_active,
                    is_low_battery,
                )
            )
        return ZoneStatusData(zones)

    def decode_ac_control(self, bytes: bytes, repeat_data_count: int) -> AcControlData:
        ac_control: list[AcControl] = []
        bits = bitarray(endian="big")

        for i in range(0, repeat_data_count):
            bits.clear()
            bits.frombytes(bytes[i * 4 : i * 4 + 4])
            # Byte 1 Bit 8-5 Power setting
            power_setting = ba2int(bits[0:4])
            if power_setting in self._set_SetPowerSetting:
                power_setting = SetPowerSetting(power_setting)
            else:
                power_setting = SetPowerSetting.KEEP_POWER_SETTING
            # Byte 1 Bit 4-1 AC number
            ac_number = ba2int(bits[4:8])
            # Byte 2 Bit 8-5 AC mode
            ac_mode = ba2int(bits[8 + 0 : 8 + 4])
            if ac_mode in self._set_SetAcMode:
                ac_mode = SetAcMode(ac_mode)
            else:
                ac_mode = SetAcMode.KEEP_AC_MODE
            # Byte 2 Bit 4-1 AC fan speed
            ac_fan_speed = ba2int(bits[8 + 4 : 8 + 8])
            if ac_fan_speed in self._set_SetAcFanSpeed:
                ac_fan_speed = SetAcFanSpeed(ac_fan_speed)
            else:
                ac_fan_speed = SetAcFanSpeed.KEEP_AC_FAN_SPEED
            # Byte 3 Setpoint control
            setpoint_control = ba2int(bits[16 : 16 + 8])
            if setpoint_control in self._set_SetpointControl:
                setpoint_control = SetpointControl(setpoint_control)
            else:
                setpoint_control = SetpointControl.INVALIDATE_DATA
            # Byte 4 Setpoint value
            setpoint = (ba2int(bits[24 : 24 + 8]) + 100) / 10

            ac_control.append(
                AcControl(
                    power_setting,
                    ac_number,
                    ac_mode,
                    ac_fan_speed,
                    setpoint_control,
                    setpoint,
                )
            )
        return AcControlData(ac_control)

    def decode_ac_status(
        self, bytes: bytes, repeat_data_count: int, repeat_data_length: int
    ) -> AcStatusData:
        if repeat_data_count == 0:
            return AcStatusData([])
        ac_status: list[AcStatus] = []
        bits = bitarray(endian="big")

        for i in range(0, repeat_data_count):
            bits.clear()
            bits.frombytes(
                bytes[
                    i * repeat_data_length : i * repeat_data_length + repeat_data_length
                ]
            )

            # Byte 1 Bit 8-5 AC power state
            ac_power_state = ba2int(bits[0:4])
            if ac_power_state in self._set_AcPowerState:
                ac_power_state = AcPowerState(ac_power_state)
            else:
                ac_power_state = AcPowerState.NOT_AVAILABLE
            # Byte 1 Bit 4-1 AC number
            ac_number = ba2int(bits[4:8])
            # Byte 2 Bit 8-5 AC mode
            ac_mode = ba2int(bits[8 + 0 : 8 + 4])
            if ac_mode in self._set_AcMode:
                ac_mode = AcMode(ac_mode)
            else:
                ac_mode = AcMode.NOT_AVAILABLE
            # Byte 2 Bit 4-1 AC fan speed
            ac_fan_speed = ba2int(bits[8 + 4 : 8 + 8])
            if ac_fan_speed in self._set_AcFanSpeed:
                ac_fan_speed = AcFanSpeed(ac_fan_speed)
            else:
                ac_fan_speed = AcFanSpeed.NOT_AVAILABLE
            # Byte 3 Setpoint
            setpoint = (ba2int(bits[16 : 16 + 8]) + 100) / 10
            # Byte 4 Bit 4 Turbo active
            turbo_active = bool(bits[24 + 4])
            # Byte 4 Bit 3 Bypass active
            bypass_active = bool(bits[24 + 5])
            # Byte 4 Bit 2 Spill active
            spill_active = bool(bits[24 + 6])
            # Byte 4 Bit 1 Timer set
            timer_set = bool(bits[24 + 7])
            # Byte 5 Bit 3-1, Byte 6 Temperature
            temperature = ba2int(bits[32 + 5 : 32 + 5 + 3 + 8])
            if temperature <= 2000:
                temperature = (temperature - 500) / 10
            else:  # Other: Not available
                temperature = None
            # Byte 7-8 Error code
            error_code = ba2int(bits[48 : 48 + 16])

            # Byte 9 NOT USED (in 1.1 docs)
            # Byte 10,11,12,13 ?????? Not in 1.1 docs, introduced in console 1.2.0

            ac_status.append(
                AcStatus(
                    ac_power_state,
                    ac_number,
                    ac_mode,
                    ac_fan_speed,
                    setpoint,
                    turbo_active,
                    bypass_active,
                    spill_active,
                    timer_set,
                    temperature,
                    error_code,
                )
            )

        return AcStatusData(ac_status)

    def decode_extended(self, bytes: bytes) -> Data:
        # Sub message type (2 bytes)
        sub_message_type = struct.unpack(">H", bytes[0:2])[0]
        message_bytes = bytes[2:]

        match sub_message_type:
            case ExtendedMessageSubType.AC_ABILITY.value:
                return self.decode_ac_ability(message_bytes)
            case ExtendedMessageSubType.AC_ERROR_INFORMATION.value:
                return self.decode_ac_error_information(message_bytes)
            case ExtendedMessageSubType.ZONE_NAME.value:
                return self.decode_zone_name(message_bytes)
            case ExtendedMessageSubType.CONSOLE_VERSION.value:
                return self.decode_console_version(message_bytes)
            case _:
                raise ValueError(f"Unknown sub message type: {hex(sub_message_type)}")

    def decode_ac_ability(self, bytes: bytes) -> Data:
        if len(bytes) == 0:
            return AcAbilityRequestData(None)
        if len(bytes) == 1:
            return AcAbilityRequestData(struct.unpack(">B", bytes[0:1])[0])

        ac_ability: list[AcAbility] = []
        bits = bitarray(endian="big")

        while len(bytes) > 2:
            # Byte 3 AC number
            ac_number = struct.unpack(">B", bytes[0:1])[0]
            # Byte 4 Following data length
            length = struct.unpack(">B", bytes[1:2])[0]
            # Byte 5-20 AC Name (Null terminated if nulls fit)
            name = bytes[2:18].decode(errors="ignore").split("\x00")[0]
            # Byte 21 start zone number
            start_zone_number = struct.unpack(">B", bytes[18:19])[0]
            # Byte 22 zone count
            zone_count = struct.unpack(">B", bytes[19:20])[0]

            bits.clear()
            bits.frombytes(bytes[20:22])
            # Byte 23 Bit 5 Cool mode
            supports_mode_cool = bool(bits[3])
            # Byte 23 Bit 4 Fan mode
            supports_mode_fan = bool(bits[4])
            # Byte 23 Bit 3 Dry mode
            supports_mode_dry = bool(bits[5])
            # Byte 23 Bit 2 Heat mode
            supports_mode_heat = bool(bits[6])
            # Byte 23 Bit 1 Auto mode
            supports_mode_auto = bool(bits[7])
            # Byte 24 Bit 8 Fan speed intelligent auto
            supports_fan_speed_intelligent_auto = bool(bits[8 + 0])
            # Byte 24 Bit 7 Fan speed turbo
            supports_fan_speed_turbo = bool(bits[8 + 1])
            # Byte 24 Bit 6 Fan speed powerful
            supports_fan_speed_powerful = bool(bits[8 + 2])
            # Byte 24 Bit 5 Fan speed high
            supports_fan_speed_high = bool(bits[8 + 3])
            # Byte 24 Bit 4 Fan speed medium
            supports_fan_speed_medium = bool(bits[8 + 4])
            # Byte 24 Bit 3 Fan speed low
            supports_fan_speed_low = bool(bits[8 + 5])
            # Byte 24 Bit 2 Fan speed quiet
            supports_fan_speed_quiet = bool(bits[8 + 6])
            # Byte 24 Bit 1 Fan speed auto
            supports_fan_speed_auto = bool(bits[8 + 7])

            # Byte 25 Min cool set point
            min_cool_set_point = struct.unpack(">B", bytes[22:23])[0]
            # Byte 26 Max cool set point
            max_cool_set_point = struct.unpack(">B", bytes[23:24])[0]
            # Byte 27 Min heat set point
            min_heat_set_point = struct.unpack(">B", bytes[24:25])[0]
            # Byte 28 Max heat set point
            max_heat_set_point = struct.unpack(">B", bytes[25:26])[0]

            ac_ability.append(
                AcAbility(
                    ac_number,
                    name,
                    start_zone_number,
                    zone_count,
                    supports_mode_cool,
                    supports_mode_fan,
                    supports_mode_dry,
                    supports_mode_heat,
                    supports_mode_auto,
                    supports_fan_speed_intelligent_auto,
                    supports_fan_speed_turbo,
                    supports_fan_speed_powerful,
                    supports_fan_speed_high,
                    supports_fan_speed_medium,
                    supports_fan_speed_low,
                    supports_fan_speed_quiet,
                    supports_fan_speed_auto,
                    min_cool_set_point,
                    max_cool_set_point,
                    min_heat_set_point,
                    max_heat_set_point,
                )
            )

            # We've now used the first 2 + length bytes, so remove them from the buffer
            bytes = bytes[2 + length :]

        return AcAbilityData(ac_ability)

    def decode_ac_error_information(self, bytes: bytes) -> Data:
        # Byte 3 AC number
        ac_number = struct.unpack(">B", bytes[0:1])[0]
        if len(bytes) == 1:
            return AcErrorInformationRequestData(ac_number)

        # Byte 4 error info length
        error_length = struct.unpack(">B", bytes[1:2])[0]
        # Byte 5-error_length error info
        error_info = bytes[2 : 2 + error_length].decode(errors="ignore")
        return AcErrorInformationData(ac_number, error_info)

    def decode_zone_name(self, bytes: bytes) -> Data:
        if len(bytes) == 0:
            return ZoneNameRequestData(None)
        if len(bytes) == 1:
            return ZoneNameRequestData(struct.unpack(">B", bytes[0:1])[0])

        names: list[ZoneName] = []

        while len(bytes) > 2:
            # Byte 3 Zone number
            zone_number = struct.unpack(">B", bytes[0:1])[0]
            # Byte 4 Name  length
            length = struct.unpack(">B", bytes[1:2])[0]
            # Byte 5..n Zone Name
            name = bytes[2 : 2 + length].decode(errors="ignore")

            names.append(ZoneName(zone_number, name))
            # We've now used the first 2 + length bytes, so remove them from the buffer
            bytes = bytes[2 + length :]

        return ZoneNameData(names)

    def decode_console_version(self, bytes: bytes) -> Data:
        if len(bytes) == 0:
            return ConsoleVersionRequestData()

        # Byte 3 Update sign (0 - latest, Other - new version)
        has_update = struct.unpack(">B", bytes[0:1])[0] != 0
        # Byte 4 version string length
        length = struct.unpack(">B", bytes[1:2])[0]
        # Byte 5..n Version string
        version = bytes[2 : 2 + length].decode(errors="ignore")

        return ConsoleVersionData(has_update, version)



class Data:
    pass


class DataPacket:
    address: int
    message_id: int
    data: Data

    def __init__(self, address: int, message_id: int, data: Data):
        self.address = address
        self.message_id = message_id
        self.data = data



from enum import Enum

from airtouch5py.packets.datapacket import Data


class AcPowerState(Enum):
    NOT_AVAILABLE = (
        0b1111  # This should be used for any value that is not explicitly defined
    )
    OFF = 0b0000
    ON = 0b0001
    AWAY_OFF = 0b0010
    AWAY_ON = 0b0011
    SLEEP = 0b0101
    # Other: Not available


class AcMode(Enum):
    NOT_AVAILABLE = (
        0b1111  # This should be used for any value that is not explicitly defined
    )
    AUTO = 0b0000
    HEAT = 0b0001
    DRY = 0b0010
    FAN = 0b0011
    COOL = 0b0100
    AUTO_HEAT = 0b1000
    AUTO_COOL = 0b1001
    # Other: Not available


class AcFanSpeed(Enum):
    NOT_AVAILABLE = (
        0b1111  # This should be used for any value that is not explicitly defined
    )
    AUTO = 0b0000
    QUIET = 0b0001
    LOW = 0b0010
    MEDIUM = 0b0011
    HIGH = 0b0100
    POWERFUL = 0b0101
    TURBO = 0b0110
    INTELLIGENT_AUTO_1 = 0b1001
    INTELLIGENT_AUTO_2 = 0b1010
    INTELLIGENT_AUTO_3 = 0b1011
    INTELLIGENT_AUTO_4 = 0b1100
    INTELLIGENT_AUTO_5 = 0b1101
    INTELLIGENT_AUTO_6 = 0b1110
    # Other: Not available


class AcStatus:
    ac_power_state: AcPowerState
    ac_number: int
    ac_mode: AcMode
    ac_fan_speed: AcFanSpeed
    ac_setpoint: float | None  # Other: Not available
    turbo_active: bool
    bypass_active: bool
    spill_active: bool
    timer_set: bool
    temperature: float | None  # Other: Not available
    error_code: int

    def __init__(
        self,
        ac_power_state: AcPowerState,
        ac_number: int,
        ac_mode: AcMode,
        ac_fan_speed: AcFanSpeed,
        ac_setpoint: float | None,
        turbo_active: bool,
        bypass_active: bool,
        spill_active: bool,
        timer_set: bool,
        temperature: float | None,
        error_code: int,
    ):
        self.ac_power_state = ac_power_state
        self.ac_number = ac_number
        self.ac_mode = ac_mode
        self.ac_fan_speed = ac_fan_speed
        self.ac_setpoint = ac_setpoint
        self.turbo_active = turbo_active
        self.bypass_active = bypass_active
        self.spill_active = spill_active
        self.timer_set = timer_set
        self.temperature = temperature
        self.error_code = error_code


class AcStatusData(Data):
    ac_status: list[AcStatus]

    def __init__(self, ac_status: list[AcStatus]):
        self.ac_status = ac_status



from airtouch5py.packets.ac_ability import AcAbilityRequestData
from airtouch5py.packets.ac_control import AcControl, AcControlData
from airtouch5py.packets.ac_error_information import AcErrorInformationRequestData
from airtouch5py.packets.ac_status import AcStatusData
from airtouch5py.packets.console_version import ConsoleVersionRequestData
from airtouch5py.packets.datapacket import DataPacket
from airtouch5py.packets.zone_control import ZoneControlData, ZoneControlZone
from airtouch5py.packets.zone_name import ZoneNameRequestData
from airtouch5py.packets.zone_status import ZoneStatusData

ADDRESS = 0x80B0
EXTENDED_ADDRESS = 0x90B0


class DataPacketFactory:
    _id: int

    def __init__(self):
        self._id = 0x01

    def zone_control(self, zones: list[ZoneControlZone]) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(ADDRESS, self._id, ZoneControlData(zones))

    def zone_status_request(self) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(ADDRESS, self._id, ZoneStatusData([]))

    def ac_control(self, ac: list[AcControl]) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(ADDRESS, self._id, AcControlData(ac))

    def ac_status_request(self) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(ADDRESS, self._id, AcStatusData([]))

    def ac_ability_request(self, ac_number: int | None = None) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(EXTENDED_ADDRESS, self._id, AcAbilityRequestData(ac_number))

    def ac_error_information_request(self, ac_number: int) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(
            EXTENDED_ADDRESS, self._id, AcErrorInformationRequestData(ac_number)
        )

    def zone_name_request(self, zone_number: int | None = None) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(EXTENDED_ADDRESS, self._id, ZoneNameRequestData(zone_number))

    def console_version_request(self) -> DataPacket:
        self._id = (self._id + 1) % 256
        return DataPacket(EXTENDED_ADDRESS, self._id, ConsoleVersionRequestData())



import logging

from airtouch5py.packet_decoder import PacketDecoder
from airtouch5py.packets.datapacket import DataPacket

_LOGGER = logging.getLogger(__name__)

# Header (4) + Address (2) + Message Id (1) + Message type (1) + Data length (2) + check bytes (2)
MINIMUM_PACKET_LENGTH = 12
HEADER = b"\x55\x55\x55\xAA"


class PacketReader:
    """
    Provide the bytes received from the airtouch 5 socket, and this class will find the packets, decode them, and return them.
    """

    _buffer: bytearray
    _packet_decoder = PacketDecoder()

    def __init__(self):
        self._buffer = bytearray()

    def read(self, data: bytes) -> list[DataPacket]:
        """
        Read the data and return a list of packets.
        """
        self._buffer.extend(data)

        packets = []

        # h. Redundant bytes in message
        # To prevent the message from containing the same data as header, a 00 is inserted after every three
        # consecutive 0x55s in the message. The inserted 00 is redundant bytes
        # ^^ In testing this actually doesn't happen. I named a zone UUUUUUU and it didn't insert any 00s

        while len(self._buffer) >= MINIMUM_PACKET_LENGTH:
            # Seek until we find the header
            while len(self._buffer) > 0 and not self._buffer.startswith(HEADER):
                self._buffer.pop(0)

            # If we don't have enough data for a packet, then we can't do anything
            if len(self._buffer) < MINIMUM_PACKET_LENGTH:
                break

            # Check if we have a full packet (Data length is now bytes 9-10)
            data_length = int.from_bytes(self._buffer[8:10], byteorder="big")
            packet_length = data_length + MINIMUM_PACKET_LENGTH
            if len(self._buffer) < packet_length:
                break

            # Decode the packet
            try:
                packet = self._packet_decoder.decode(self._buffer[:packet_length])
                packets.append(packet)
            except Exception as e:
                _LOGGER.debug(f"Error decoding packet: {e}")

            # remove the packet from the buffer
            self._buffer = self._buffer[packet_length:]

        return packets



from airtouch5py.packets.datapacket import Data


class ConsoleVersionRequestData(Data):
    pass


class ConsoleVersionData(Data):
    has_update: bool
    version: str

    def __init__(self, has_update: bool, version: str):
        self.has_update = has_update
        self.version = version



from enum import Enum

from airtouch5py.packets.datapacket import Data


class ZoneSettingValue(Enum):
    KEEP_SETTING_VALUE = (
        0b000  # This should be used for any value that is not explicitly defined
    )
    VALUE_DECREASE = 0b010
    VALUE_INCREASE = 0b011
    SET_OPEN_PERCENTAGE = 0b100
    SET_TARGET_SETPOINT = 0b101
    # Other: Keep setting value


class ZoneSettingPower(Enum):
    KEEP_POWER_STATE = (
        0b000  # This should be used for any value that is not explicitly defined
    )
    CHANGE_ON_OFF_STATE = 0b001
    SET_TO_OFF = 0b010
    SET_TO_ON = 0b011
    SET_TO_TURBO = 0b101
    # Other: Keep power state


class ZoneControlZone:
    zone_number: int
    zone_setting_value: ZoneSettingValue
    power: ZoneSettingPower
    value_to_set: float

    def __init__(
        self,
        zone_number: int,
        zone_setting_value: ZoneSettingValue,
        power: ZoneSettingPower,
        value_to_set: float,
    ):
        self.zone_number = zone_number
        self.zone_setting_value = zone_setting_value
        self.power = power
        self.value_to_set = value_to_set


class ZoneControlData(Data):
    """
    Zone control messages are to control all zones. Each message to AirTouch is to control one or more specific zones.
    AirTouch will respond a message with sub type 0x21. (Zone status message)
    """

    zones: list[ZoneControlZone]

    def __init__(self, zones: list[ZoneControlZone]):
        self.zones = zones



from enum import Enum

from airtouch5py.packets.datapacket import Data


class ZonePowerState(Enum):
    OFF = 0b00
    ON = 0b01
    TURBO = 0b11


class ControlMethod(Enum):
    TEMPERATURE_CONTROL = 1
    PERCENTAGE_CONTROL = 0


class ZoneStatusZone:
    zone_power_state: ZonePowerState
    zone_number: int
    control_method: ControlMethod
    open_percentage: float  # 0.0 - 1.0
    set_point: float | None  # 0xFF invalid (None)
    has_sensor: bool
    temperature: float | None  # Temperature > 150 invalid (None)
    spill_active: bool
    is_low_battery: bool

    def __init__(
        self,
        zone_power_state: ZonePowerState,
        zone_number: int,
        control_method: ControlMethod,
        open_percentage: float,
        set_point: float | None,
        has_sensor: bool,
        temperature: float | None,
        spill_active: bool,
        is_low_battery: bool,
    ):
        self.zone_power_state = zone_power_state
        self.zone_number = zone_number
        self.control_method = control_method
        self.open_percentage = open_percentage
        self.set_point = set_point
        self.has_sensor = has_sensor
        self.temperature = temperature
        self.spill_active = spill_active
        self.is_low_battery = is_low_battery


class ZoneStatusData(Data):
    """
    Send this message to AirTouch without any sub data (data length: 0x00 0x08, repeat count: 0x00, repeat length: 0x00) to request zone status from AirTouch.
    Note: AirTouch will send a zone status message automatically when zone status is changed.
    """

    zones: list[ZoneStatusZone]

    def __init__(self, zones: list[ZoneStatusZone]):
        self.zones = zones



from enum import Enum


class MessageType(Enum):
    # Control command and status message
    CONTROL_STATUS = 0xC0
    EXTENDED = 0x1F


class ControlStatusSubType(Enum):
    ZONE_CONTROL = 0x20
    ZONE_STATUS = 0x21
    AC_CONTROL = 0x22
    AC_STATUS = 0x23


class ExtendedMessageSubType(Enum):
    AC_ABILITY = 0xFF11
    AC_ERROR_INFORMATION = 0xFF10
    ZONE_NAME = 0xFF13
    CONSOLE_VERSION = 0xFF30



from enum import Enum

from airtouch5py.packets.datapacket import Data


class SetPowerSetting(Enum):
    KEEP_POWER_SETTING = (
        0b0000  # This should be used for any value that is not explicitly defined
    )
    CHANGE_ON_OFF_STATUS = 0b0001  # Does this mean toggle?
    SET_TO_OFF = 0b0010
    SET_TO_ON = 0b0011
    SET_TO_AWAY = 0b0100
    SET_TO_SLEEP = 0b0101
    # Other: Keep power setting


class SetAcMode(Enum):
    KEEP_AC_MODE = (
        0b1111  # This should be used for any value that is not explicitly defined
    )
    SET_TO_AUTO = 0b0000
    SET_TO_HEAT = 0b0001
    SET_TO_DRY = 0b0010
    SET_TO_FAN = 0b0011
    SET_TO_COOL = 0b0100
    # Other: keep mode setting


class SetAcFanSpeed(Enum):
    KEEP_AC_FAN_SPEED = (
        0b1111  # This should be used for any value that is not explicitly defined
    )
    SET_TO_AUTO = 0b0000
    SET_TO_QUIET = 0b0001  # "Quite" in the docs
    SET_TO_LOW = 0b0010
    SET_TO_MEDIUM = 0b0011
    SET_TO_HIGH = 0b0100
    SET_TO_POWERFUL = 0b0101
    SET_TO_TURBO = 0b0110
    SET_TO_INTELLIGENT_AUTO = 0b1000
    # Other: keep fan speed setting


class SetpointControl(Enum):
    INVALIDATE_DATA = (
        0xFF  # This should be used for any value that is not explicitly defined
    )
    CHANGE_SETPOINT = 0x40
    KEEP_SETPOINT_VALUE = 0x00
    # Other: Invalidate data (????)


class AcControl:
    power_setting: SetPowerSetting
    ac_number: int
    ac_mode: SetAcMode
    ac_fan_speed: SetAcFanSpeed
    setpoint_control: SetpointControl
    setpoint: float

    def __init__(
        self,
        power_setting: SetPowerSetting,
        ac_number: int,
        ac_mode: SetAcMode,
        ac_fan_speed: SetAcFanSpeed,
        setpoint_control: SetpointControl,
        setpoint: float,
    ):
        self.power_setting = power_setting
        self.ac_number = ac_number
        self.ac_mode = ac_mode
        self.ac_fan_speed = ac_fan_speed
        self.setpoint_control = setpoint_control
        self.setpoint = setpoint


class AcControlData(Data):
    """
    AC control messages are to control all ACs. Each message to AirTouch is to control one or more specific ACs.
    AirTouch will respond a message with sub type 0x23. (AC status message)
    """

    ac_control: list[AcControl]

    def __init__(self, ac_control: list[AcControl]):
        self.ac_control = ac_control



from airtouch5py.packets.datapacket import Data


class ZoneName:
    zone_number: int
    zone_name: str

    def __init__(self, zone_number: int, zone_name: str):
        self.zone_number = zone_number
        self.zone_name = zone_name


class ZoneNameRequestData(Data):
    zone_number: int | None  # None for all

    def __init__(self, zone_number: int | None):
        self.zone_number = zone_number


class ZoneNameData(Data):
    zone_names: list[ZoneName]

    def __init__(self, zone_names: list[ZoneName]):
        self.zone_names = zone_names



import struct

from airtouch5py.packet_fields import MessageType
from airtouch5py.packets.ac_ability import AcAbilityData, AcAbilityRequestData
from airtouch5py.packets.ac_control import AcControlData, SetpointControl
from airtouch5py.packets.ac_error_information import (
    AcErrorInformationData,
    AcErrorInformationRequestData,
)
from airtouch5py.packets.ac_status import AcStatusData
from airtouch5py.packets.console_version import (
    ConsoleVersionData,
    ConsoleVersionRequestData,
)
from airtouch5py.packets.datapacket import Data, DataPacket
from airtouch5py.packets.zone_control import ZoneControlData, ZoneSettingValue
from airtouch5py.packets.zone_name import ZoneNameData, ZoneNameRequestData

from airtouch5py.packets.zone_status import ZoneStatusData

from crc import Calculator, Crc16

_calculator = Calculator(Crc16.MODBUS)  # type: ignore


class PacketEncoder:
    header = b"\x55\x55\x55\xAA"

    def encode(self, packet: DataPacket) -> bytes:
        # Header is always 0x55 0x55 0x55 0xAA
        res = self.header
        # Address, Message id
        res += struct.pack(">HB", packet.address, packet.message_id)
        # Message type
        res += struct.pack(">B", self._message_type(packet.data).value)

        # Data length (2bytes)
        packet_data = self._encode_data(packet.data)
        res += struct.pack(">H", len(packet_data))
        # Data
        res += packet_data

        # CRC16 check bytes
        res += struct.pack(">H", _calculator.checksum(res[4:]))

        # h. Redundant bytes in message
        # To prevent the message from containing the same data as header, a 00 is inserted after every three
        # consecutive 0x55s in the message. The inserted 00 is redundant bytes
        # ^^ In testing this actually doesn't happen. I named a zone UUUUUUU and it didn't insert any 00s

        return res

    def _message_type(self, data: Data) -> MessageType:
        match data:
            case ZoneControlData() | ZoneStatusData() | AcControlData() | AcStatusData():
                return MessageType.CONTROL_STATUS
            case AcAbilityRequestData() | AcAbilityData() | AcErrorInformationRequestData() | AcErrorInformationData() | ZoneNameRequestData() | ZoneNameData() | ConsoleVersionRequestData() | ConsoleVersionData():
                return MessageType.EXTENDED
            case _:
                raise Exception(f"Unknown message type for {data.__class__.__name__}")

    def _encode_data(self, data: Data) -> bytes:
        match data:
            case ZoneControlData():
                return self._encode_zone_control_data(data)
            case ZoneStatusData():
                return self._encode_zone_status_data(data)
            case AcControlData():
                return self._encode_ac_control_data(data)
            case AcStatusData():
                return self._encode_ac_status_data(data)
            case AcAbilityRequestData():
                return self._encode_ac_ability_request_data(data)
            case AcAbilityData():
                return self._encode_ac_ability_data(data)
            case AcErrorInformationRequestData():
                return self._encode_ac_error_information_request_data(data)
            case AcErrorInformationData():
                return self._encode_ac_error_information_data(data)
            case ZoneNameRequestData():
                return self._encode_zone_name_request_data(data)
            case ZoneNameData():
                return self._encode_zone_name_data(data)
            case ConsoleVersionRequestData():
                return self._encode_console_version_request_data(data)
            case ConsoleVersionData():
                return self._encode_console_version_data(data)
            case _:
                raise Exception(f"Unknown data type for {data.__class__.__name__}")

    def _encode_zone_control_data(self, data: ZoneControlData) -> bytes:
        # Sub message type 0x20
        # 0
        # no normal data (2 bytes)
        # each repeat data length (2 bytes, 4)
        res = b"\x20\x00\x00\x00\x00\x04"
        # repeat count (2 bytes)
        res += struct.pack(">H", len(data.zones))

        # pack the zones
        for zone in data.zones:
            # Byte 1 Bit 8-7 Keep 0
            # Byte 1 Bit 6-1 Zone number
            res += struct.pack(">B", zone.zone_number)

            # Byte 2 Bit 8-7 Zone setting value
            # Byte 2 Bit 5-4 Keep 0
            # Byte 2 Bit 3-1 Power
            res += struct.pack(
                ">B", (zone.zone_setting_value.value << 5) | (zone.power.value << 0)
            )

            # Byte 3 Value to Set
            if zone.zone_setting_value == ZoneSettingValue.SET_OPEN_PERCENTAGE:
                res += struct.pack(">B", int(zone.value_to_set * 100))
            elif zone.zone_setting_value == ZoneSettingValue.SET_TARGET_SETPOINT:
                res += struct.pack(">B", int(zone.value_to_set * 10 - 100))
            elif zone.zone_setting_value == ZoneSettingValue.KEEP_SETTING_VALUE:
                res += b"\xFF"
            else:
                raise Exception(f"Unknown zone setting value {zone.zone_setting_value}")

            # Byte 4 Keep 0
            res += b"\x00"

        return res

    def _encode_zone_status_data(self, data: ZoneStatusData) -> bytes:
        # Sub message type 0x21
        # 0
        # no normal data (2 bytes)
        # each repeat data length, only if there is repeat data (2 bytes, 8)
        res = b"\x21\x00\x00\x00\x00"
        if len(data.zones) == 0:
            res += b"\x00"
        else:
            res += b"\x08"
        # repeat count (2 bytes)
        res += struct.pack(">H", len(data.zones))

        # pack the zone statuses
        for zone in data.zones:
            # Byte 1 Bit 8-7 Zone power state
            # Byte 1 Bit 6-1 Zone number
            res += struct.pack(
                ">B", (zone.zone_power_state.value << 6) | (zone.zone_number)
            )

            # Byte 2 Bit 8 control method (temperature = 1, percentage = 0)
            # Byte 2 Bit 7-1 Open percentage
            res += struct.pack(
                ">B", (zone.control_method.value << 7) | int(zone.open_percentage * 100)
            )

            # Byte 3 Set point setpoint=(value+100)/10, 0xFF invalid (None)
            if zone.set_point is None:
                res += b"\xFF"
            else:
                res += struct.pack(">B", int(zone.set_point * 10 - 100))

            # Byte 4 Bit 8 has sensor
            # Byte 4 Bit 7-1 NOT USED
            res += struct.pack(">B", (zone.has_sensor << 7))

            # Byte 5-6 Temperature Temperature > 150 invalid (None)
            if zone.temperature is None:
                res += b"\x07\xFF"  # The sample packs this
            else:
                res += struct.pack(">H", int(zone.temperature * 10 + 500))

            # Byte 7 Bit 8-3 NOT USED
            # Byte 7 Bit 2 Spill active
            # Byte 7 Bit 1 Low battery
            res += struct.pack(
                ">B", (zone.spill_active << 1) | (zone.is_low_battery << 0)
            )

            # Byte 8 NOT USED
            res += b"\x00"

        return res

    def _encode_ac_control_data(self, data: AcControlData) -> bytes:
        # Sub message type 0x22
        # 0
        # no normal data (2 bytes)
        # each repeat data length (2 bytes, 4)
        res = b"\x22\x00\x00\x00\x00\x04"
        # repeat count (2 bytes)
        res += struct.pack(">H", len(data.ac_control))

        # pack the ac controls
        for ac in data.ac_control:
            # Byte 1 Bit 8-5 Power setting
            # Byte 1 Bit 4-1 AC number
            res += struct.pack(">B", (ac.power_setting.value << 4) | (ac.ac_number))

            # Byte 2 Bit 8-5 AC mode
            # Byte 2 Bit 4-1 AC fan speed
            res += struct.pack(">B", (ac.ac_mode.value << 4) | (ac.ac_fan_speed.value))

            # Byte 3 Setpoint control
            res += struct.pack(">B", ac.setpoint_control.value)

            # Byte 4 Setpoint value (Available when byte 3 is Change setpoint)
            match ac.setpoint_control:
                case SetpointControl.CHANGE_SETPOINT:
                    res += struct.pack(">B", int(ac.setpoint * 10 - 100))
                case SetpointControl.KEEP_SETPOINT_VALUE | SetpointControl.INVALIDATE_DATA:
                    res += b"\xFF"
                case _:
                    raise Exception(
                        f"Unsupported setpoint control {ac.setpoint_control}"
                    )

        return res

    def _encode_ac_status_data(self, data: AcStatusData) -> bytes:
        # Sub message type 0x23
        # 0
        # no normal data (2 bytes)
        # each repeat data length, only if there is repeat data (2 bytes, 10)
        res = b"\x23\x00\x00\x00\x00"
        if len(data.ac_status) == 0:
            res += b"\x00"
        else:
            res += b"\x0A"
        # repeat count (2 bytes)
        res += struct.pack(">H", len(data.ac_status))

        # pack the ac statuses
        for ac in data.ac_status:
            # Byte 1 Bit 8-5 AC power state
            # Byte 1 Bit 4-1 AC number
            res += struct.pack(">B", (ac.ac_power_state.value << 4) | (ac.ac_number))

            # Byte 2 Bit 8-5 AC mode
            # Byte 2 Bit 4-1 AC fan speed
            res += struct.pack(">B", (ac.ac_mode.value << 4) | (ac.ac_fan_speed.value))

            # Byte 3 Setpoint (0xFF if Not Available)
            if ac.ac_setpoint is None:
                res += b"\xFF"
            else:
                res += struct.pack(">B", int(ac.ac_setpoint * 10 - 100))

            # Byte 4 Bit 8-5 NOT USED (Packed as C in the sample)
            # Byte 4 Bit 4 Turbo active
            # Byte 4 Bit 3 Bypass active
            # Byte 4 Bit 2 Spill active
            # Byte 4 Bit 1 Timer set
            res += struct.pack(
                ">B",
                0xC0
                | ac.turbo_active << 3
                | ac.bypass_active << 2
                | ac.spill_active << 1
                | ac.timer_set << 0,
            )

            # Byte 5-6 Temperature (0x07FF if Not Available)
            if ac.temperature is None:
                res += b"\x07\xFF"
            else:
                res += struct.pack(">H", int(ac.temperature * 10 + 500))

            # Byte 7-8 Error code
            res += struct.pack(">H", ac.error_code)

            # Byte 9-10 NOT USED (Sample packs 8000)
            res += b"\x80\x00"

        return res

    def _encode_ac_ability_request_data(self, data: AcAbilityRequestData) -> bytes:
        # Sub message type 0xFF11
        res = b"\xFF\x11"

        if data.ac_number is not None:
            res += struct.pack(">B", data.ac_number)

        return res

    def _encode_ac_ability_data(self, data: AcAbilityData) -> bytes:
        # Sub message type 0xFF11
        res = b"\xFF\x11"

        # Pack each ac
        for ac in data.ac_ability:
            # Byte 3 AC number
            res += struct.pack(">B", ac.ac_number)

            # Byte 4 Following data length (24)
            res += b"\x18"

            # Byte 5-20 AC Name (Null padded)
            res += ac.ac_name.encode("ascii").ljust(16, b"\x00")

            # Byte 21 start zone number
            res += struct.pack(">B", ac.start_zone_number)

            # Byte 22 zone count
            res += struct.pack(">B", ac.zone_count)

            # Byte 23 Bit 8-6 NOT USED
            # Byte 23 Bit 5 Cool mode supported
            # Byte 23 Bit 4 Fan mode supported
            # Byte 23 Bit 3 Dry mode supported
            # Byte 23 Bit 2 Heat mode supported
            # Byte 23 Bit 1 Auto mode supported
            res += struct.pack(
                ">B",
                0x00
                | ac.supports_mode_cool << 4
                | ac.supports_mode_fan << 3
                | ac.supports_mode_dry << 2
                | ac.supports_mode_heat << 1
                | ac.supports_mode_auto << 0,
            )

            # Byte 24 Bit 8 Supports intelligent auto fan speed
            # Byte 24 Bit 7 Supports turbo fan speed
            # Byte 24 Bit 6 Supports powerful fan speed
            # Byte 24 Bit 5 Supports high fan speed
            # Byte 24 Bit 4 Supports medium fan speed
            # Byte 24 Bit 3 Supports low fan speed
            # Byte 24 Bit 2 Supports quiet fan speed
            # Byte 24 Bit 1 Supports auto fan speed
            res += struct.pack(
                ">B",
                0x00
                | ac.supports_fan_speed_intelligent_auto << 7
                | ac.supports_fan_speed_turbo << 6
                | ac.supports_fan_speed_powerful << 5
                | ac.supports_fan_speed_high << 4
                | ac.supports_fan_speed_medium << 3
                | ac.supports_fan_speed_low << 2
                | ac.supports_fan_speed_quiet << 1
                | ac.supports_fan_speed_auto << 0,
            )

            # Byte 25 Min cool set point
            res += struct.pack(">B", ac.min_cool_set_point)
            # Byte 26 Max cool set point
            res += struct.pack(">B", ac.max_cool_set_point)
            # Byte 27 Min heat set point
            res += struct.pack(">B", ac.min_heat_set_point)
            # Byte 28 Max heat set point
            res += struct.pack(">B", ac.max_heat_set_point)

        return res

    def _encode_ac_error_information_request_data(
        self, data: AcErrorInformationRequestData
    ) -> bytes:
        # Sub message type 0xFF10
        res = b"\xFF\x10"

        if data.ac_number is not None:
            res += struct.pack(">B", data.ac_number)

        return res

    def _encode_ac_error_information_data(self, data: AcErrorInformationData) -> bytes:
        # Sub message type 0xFF10
        res = b"\xFF\x10"

        # Byte 3 AC number
        res += struct.pack(">B", data.ac_number)
        # Byte 4 Error info length (0 if no error)
        res += struct.pack(">B", len(data.error_info))
        # Byte 5.. Error info (no bytes if no error, no null at end)
        res += data.error_info.encode("ascii")

        return res

    def _encode_zone_name_request_data(self, data: ZoneNameRequestData) -> bytes:
        # Sub message type 0xFF13
        res = b"\xFF\x13"

        if data.zone_number is not None:
            res += struct.pack(">B", data.zone_number)

        return res

    def _encode_zone_name_data(self, data: ZoneNameData) -> bytes:
        # Sub message type 0xFF13
        res = b"\xFF\x13"

        # Pack each zone
        for zone in data.zone_names:
            # Byte 3 Zone number
            res += struct.pack(">B", zone.zone_number)

            # Byte 4 Name length
            res += struct.pack(">B", len(zone.zone_name))

            # Byte 5.. Zone name (no null at end)
            res += zone.zone_name.encode("ascii")

        return res

    def _encode_console_version_request_data(
        self, data: ConsoleVersionRequestData
    ) -> bytes:
        # Sub message type 0xFF30
        res = b"\xFF\x30"

        return res

    def _encode_console_version_data(self, data: ConsoleVersionData) -> bytes:
        # Sub message type 0xFF30
        res = b"\xFF\x30"

        # Byte 3 Update sign (0 - latest, Other - new version available)
        res += struct.pack(">B", data.has_update)

        # Byte 4 Version length
        res += struct.pack(">B", len(data.version))

        # Byte 5.. Version (no null at end)
        res += data.version.encode("ascii")

        return res
